package envoy.client;

import java.io.File;
import java.io.IOException;
import java.util.*;

import envoy.data.User;
import envoy.util.SerializationUtils;

/**
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>LocalDB.java</strong><br>
 * Created: <strong>27.10.2019</strong><br>
 *
 * @author Kai S. K. Engelbart
 * @author Maximilian K&auml;fer
 * @since Envoy v0.1-alpha
 */
public class LocalDB {

	private File				localDBDir, localDBFile, usersFile;
	private User				user;
	private Map<String, User>	users	= new HashMap<>();
	private List<Chat>			chats	= new ArrayList<>();

	/**
	 * Constructs an empty local database. To serialize any chats to the file
	 * system, call {@link LocalDB#initializeDBFile()}.
	 *
	 * @param localDBDir the directory in which to store users and chats
	 * @throws IOException if the LocalDB could not be initialized
	 * @since Envoy v0.1-alpha
	 */
	public LocalDB(File localDBDir) throws IOException {
		this.localDBDir = localDBDir;

		// Initialize local database directory
		if (localDBDir.exists() && !localDBDir.isDirectory())
			throw new IOException(String.format("LocalDBDir '%s' is not a directory!", localDBDir.getAbsolutePath()));
		usersFile = new File(localDBDir, "users.db");
	}

	/**
	 * Creates a database file for a user-specific list of chats.
	 *
	 * @throws NullPointerException if the client user is not yet specified
	 * @since Envoy v0.1-alpha
	 */
	public void initializeDBFile() {
		if (user == null) throw new NullPointerException("Client user is null");
		localDBFile = new File(localDBDir, user.getId() + ".db");
	}

	/**
	 * Stores all users to the local database. If the client user is specified, the
	 * chats related to this user are stored as well.
	 *
	 * @throws IOException if something went wrong during saving
	 * @since Envoy v0.1-alpha
	 */
	public void save() throws IOException {
		// Save users
		SerializationUtils.write(usersFile, users);

		// Save chats
		SerializationUtils.write(localDBFile, chats);
	}

	/**
	 * Loads all users that are stored in the local database.
	 * 
	 * @throws IOException            if the loading process failed
	 * @throws ClassNotFoundException if the loading process failed
	 * @since Envoy v0.2-alpha
	 */
	public void loadUsers() throws ClassNotFoundException, IOException { users = SerializationUtils.read(usersFile, HashMap.class); }

	/**
	 * Loads all chats saved by Envoy for the client user.
	 * 
	 * @throws IOException            if the loading process failed
	 * @throws ClassNotFoundException if the loading process failed
	 * @since Envoy v0.1-alpha
	 */
	public void loadChats() throws ClassNotFoundException, IOException { chats = SerializationUtils.read(localDBFile, ArrayList.class); }

	/**
	 * @return a {@code Map<String, User>} of all users stored locally with their
	 *         user names as keys
	 * @since Envoy v0.2-alpha
	 */
	public Map<String, User> getUsers() { return users; }

	/**
	 * @param users the users to set
	 */
	public void setUsers(Map<String, User> users) { this.users = users; }

	/**
	 * @return all saved {@link Chat} objects that list the client user as the
	 *         sender
	 * @since Envoy v0.1-alpha
	 **/
	public List<Chat> getChats() { return chats; }

	/**
	 * @param chats the chats to set
	 */
	public void setChats(List<Chat> chats) { this.chats = chats; }

	/**
	 * @return the {@link User} who initialized the local database
	 * @since Envoy v0.2-alpha
	 */
	public User getUser() { return user; }

	/**
	 * @param user the user to set
	 * @since Envoy v0.2-alpha
	 */
	public void setUser(User user) { this.user = user; }
}